#include <gtk/gtk.h>
#include "sokodan.h"
#include "guix.h"
#include "gui.h"
#include "pixmaps.h"

void launch_solve (void); /* in progress_dialog.c */
static void draw_case   (GtkWidget *pix_wid, Case value, gboolean is_player_here);
static void change_grid (GtkWidget *pix_wid, Coordinate x, Coordinate y);

void gui_draw_down (GtkWidget *eventbox,
                    GdkEventButton *event)
{
  GtkWidget *pix_wid;
  Coordinate x, y;

	if (event->type != GDK_BUTTON_PRESS
			|| event->button != 1)
		return;

  pix_wid = GTK_BIN (eventbox)->child;
  x = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT(pix_wid), "x") );
  y = GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT(pix_wid), "y") );

  switch (Gui_State.tool)
  {
  case TOOL_GEM:
    Gui_State.tool_draws = ! CONTAIN (GEM, Problem.grid[ x ][ y ]);
		Gui_State.tool_drawing = TRUE;
    change_grid (pix_wid, x, y);
    break;

  case TOOL_TARGET:
    Gui_State.tool_draws = ! CONTAIN (TARGET, Problem.grid[ x ][ y ]);
		Gui_State.tool_drawing = TRUE;
    change_grid (pix_wid, x, y);
    break;

  case TOOL_WALL:
    Gui_State.tool_draws = ! CONTAIN (WALL, Problem.grid[ x ][ y ]);
		Gui_State.tool_drawing = TRUE;
    change_grid (pix_wid, x, y);
    break;

  case TOOL_BACKGROUND:
		Gui_State.tool_drawing = TRUE;
    change_grid (pix_wid, x, y);
    break;

  case TOOL_PLAYER:
    {
      Coordinate px, py;
      
      if  ((px = Problem.player_x) < WIDTH)
        /* There was already a player somewhere */
      {
        py = Problem.player_y;
        g_assert (py < HEIGHT);

        /* Remove the player there */
        draw_case (Gui_Elements.pixmap_table[ px ][ py ],
                   Problem.grid[ px ][ py ],
                   FALSE);
      }

      /* and put it here. Remove wall and gem, if any */
      REMOVE (WALL | GEM, Problem.grid[ x ][ y ]);
      Problem.player_x = x;
      Problem.player_y = y;

      draw_case (pix_wid,
                 Problem.grid[ x ][ y ],
                 TRUE);
    }
    break;

  default:
    g_assert_not_reached ();
  }
}


void gui_draw_enter (GtkWidget *eventbox)
{
  if (Gui_State.tool_drawing)
  {
    GtkWidget *pix_wid = GTK_BIN (eventbox)->child;

    change_grid (pix_wid,
                 GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT(pix_wid), "x")),
                 GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT(pix_wid), "y")) );
  }
}


void gui_draw_up (GtkWidget *eventbox)
{
  Gui_State.tool_drawing = FALSE;
}


void gui_tool_clicked (GtkWidget *button,
                       gpointer   which_tool)
{
  Gui_State.tool = (ToolType) GPOINTER_TO_INT (which_tool);
}


void gui_clear (GtkWidget *widget)
{
  gint x, y;

  for (y = 0; y < HEIGHT; y++)
    for (x = 0; x < WIDTH; x++)
    {
      gtk_pixmap_set (GTK_PIXMAP (Gui_Elements.pixmap_table[ x ][ y ]),
                      Case_Pixmap[ BACKGROUND ], NULL);
      gtk_object_set_data (GTK_OBJECT (Gui_Elements.pixmap_table[ x ][ y ]),
                           "value", GINT_TO_POINTER (BACKGROUND));
      Problem.grid[ x ][ y ] = BACKGROUND;
    }
  
  Problem.player_x = WIDTH;
  Problem.player_y = HEIGHT;
}


void gui_solve (GtkWidget *widget)
{
  /* Check 1: make sure there's a player */
  if (Problem.player_x == WIDTH)
  {
    g_assert (Problem.player_y == HEIGHT);

    gui_info_show (_("Problme incomplet"),
                   _("Le problme n'est pas complet : il n'y a pas de joueur dans la grille.\n"
                     "Veuillez cliquer sur le bouton reprsentant un joueur, \n"
                     "puis sur la grille pour le placer."));
    return;
  }

  /* Check 2: make sure there are as many gems than there are targets */
  /* Check also that there's at least one gem not on a target */
  {
    gint difference;
    gboolean is_stupid;
    Coordinate x, y;

    difference = 0;
    is_stupid = TRUE; /* default ;-) */

    for (x = 0; x < WIDTH; x++)
      for (y = 0; y < HEIGHT; y++)
      {
        if (CONTAIN (TARGET, Problem.grid[ x ][ y ]))
          difference++;

        if (CONTAIN (GEM, Problem.grid[ x ][ y ]))
          difference--;

        if (is_stupid && difference != 0)
          is_stupid = FALSE;
      }

    if (difference != 0)
    {
      gui_info_show (_("Problme incohrent"),
                     _("Le problme n'est pas cohrent : il n'y a pas autant de pierres que de cibles.\n"
                       "Veuillez modifier le problme et corriger cela."));
      return;
    }

    if (is_stupid)
    {
      gui_info_show (_("Problme trivial"),
                     _("Le problme ... n'en est pas un.\n"
                       "Je ne vais pas utiliser du temps CPU pour a.\n"
                       "Veuillez modifier le problme pour qu'il soit un peu plus intressant."));
      return;
    }
  }

  launch_solve ();
}


static void draw_case (GtkWidget *pix_wid,
                       Case       value,
                       gboolean   is_player_here)
{
  int index = value + (is_player_here ? 1 : 0);

  gtk_pixmap_set (GTK_PIXMAP (pix_wid), Case_Pixmap[ index ], NULL);
  gtk_object_set_data (GTK_OBJECT (pix_wid), "value", GINT_TO_POINTER (index));
}


static void change_grid (GtkWidget *pix_wid,
                         Coordinate x,
                         Coordinate y)
{
  switch (Gui_State.tool)
  {
  case TOOL_GEM:
    /* When tool_draws, a gem is added unless there is
     * a wall or the player. If !tool_draws, we always
     * remove the gem. */

    if (Gui_State.tool_draws)
    {
      if ( !( CONTAIN (WALL, Problem.grid[ x ][ y ])
							||
              (x    == Problem.player_x
               && y == Problem.player_y) ) )

        ADD (GEM, Problem.grid[ x ][ y ]);
    }
    else
    {
      REMOVE (GEM, Problem.grid[ x ][ y ]);
    }
 
    break;

  case TOOL_TARGET:
    /* When tool_draws, a target is added unless on
     * a wall. If !tool_draws, we always remove the target */

    if (Gui_State.tool_draws)
    {
      if ( ! CONTAIN (WALL, Problem.grid[ x ][ y ]) )
        ADD (TARGET, Problem.grid[ x ][ y ]);
    }
    else
    {
      REMOVE (TARGET, Problem.grid[ x ][ y ]);
    }

    break;

  case TOOL_WALL:
    /* When tool_draws, a wall is added unless on the player.
     * Any target or gem is removed. If !tool_draws, we
     * just remove the wall (if any) */

    if (Gui_State.tool_draws)
    {
      if ( !(x    == Problem.player_x
             && y == Problem.player_y) )
      {
        ADD (WALL, Problem.grid[ x ][ y ]);
        REMOVE (GEM | TARGET, Problem.grid[ x ][ y ]);
      }
    }
    else
    {
      REMOVE (WALL, Problem.grid[ x ][ y ]);
    }

    break;

  case TOOL_BACKGROUND:
    /* tool_draws doesn't apply here. We always set
     * the case to background. */

    Problem.grid[ x ][ y ] = BACKGROUND;

    break;

  default:
    g_assert_not_reached ();
  }

  draw_case (pix_wid,
             Problem.grid[ x ][ y ],
             (x    == Problem.player_x
              && y == Problem.player_y) );
}
